# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""DoJSON geo utils."""

from __future__ import absolute_import, division, print_function

import six

from inspire_utils.helpers import force_list


country_to_iso_code = {
    'AFGHANISTAN': 'AF',
    'ÅLAND ISLANDS': 'AX',
    'ALBANIA': 'AL',
    'ALGERIA': 'DZ',
    'AMERICAN SAMOA': 'AS',
    'ANDORRA': 'AD',
    'ANGOLA': 'AO',
    'ANGUILLA': 'AI',
    'ANTARCTICA': 'AQ',
    'ANTIGUA AND BARBUDA': 'AG',
    'ARGENTINA': 'AR',
    'ARMENIA': 'AM',
    'ARUBA': 'AW',
    'AUSTRALIA': 'AU',
    'AUSTRIA': 'AT',
    'AZERBAIJAN': 'AZ',
    'BAHAMAS': 'BS',
    'BAHRAIN': 'BH',
    'BANGLADESH': 'BD',
    'BARBADOS': 'BB',
    'BELARUS': 'BY',
    'BELGIUM': 'BE',
    'BELIZE': 'BZ',
    'BENIN': 'BJ',
    'BERMUDA': 'BM',
    'BHUTAN': 'BT',
    'BOLIVIA, PLURINATIONAL STATE OF': 'BO',
    'BONAIRE, SINT EUSTATIUS AND SABA': 'BQ',
    'BOSNIA AND HERZEGOVINA': 'BA',
    'BOTSWANA': 'BW',
    'BOUVET ISLAND': 'BV',
    'BRAZIL': 'BR',
    'BRITISH INDIAN OCEAN TERRITORY': 'IO',
    'BRUNEI DARUSSALAM': 'BN',
    'BULGARIA': 'BG',
    'BURKINA FASO': 'BF',
    'BURUNDI': 'BI',
    'CAMBODIA': 'KH',
    'CAMEROON': 'CM',
    'CANADA': 'CA',
    'CAPE VERDE': 'CV',
    'CAYMAN ISLANDS': 'KY',
    'CENTRAL AFRICAN REPUBLIC': 'CF',
    'CHAD': 'TD',
    'CHILE': 'CL',
    'CHINA': 'CN',
    'CHRISTMAS ISLAND': 'CX',
    'COCOS (KEELING) ISLANDS': 'CC',
    'COLOMBIA': 'CO',
    'COMOROS': 'KM',
    'CONGO': 'CG',
    'CONGO, THE DEMOCRATIC REPUBLIC OF THE': 'CD',
    'COOK ISLANDS': 'CK',
    'COSTA RICA': 'CR',
    'CÔTE D\'IVOIRE': 'CI',
    'CROATIA': 'HR',
    'CZECHOSLOVAKIA': 'CS',
    'CUBA': 'CU',
    'CURAÇAO': 'CW',
    'CYPRUS': 'CY',
    'CZECH REPUBLIC': 'CZ',
    'DENMARK': 'DK',
    'DJIBOUTI': 'DJ',
    'DOMINICA': 'DM',
    'DOMINICAN REPUBLIC': 'DO',
    'ECUADOR': 'EC',
    'EGYPT': 'EG',
    'EL SALVADOR': 'SV',
    'EQUATORIAL GUINEA': 'GQ',
    'ERITREA': 'ER',
    'ESTONIA': 'EE',
    'ETHIOPIA': 'ET',
    'FALKLAND ISLANDS (MALVINAS)': 'FK',
    'FAROE ISLANDS': 'FO',
    'FIJI': 'FJ',
    'FINLAND': 'FI',
    'FRANCE': 'FR',
    'FRENCH GUIANA': 'GF',
    'FRENCH POLYNESIA': 'PF',
    'FRENCH SOUTHERN TERRITORIES': 'TF',
    'GABON': 'GA',
    'GAMBIA': 'GM',
    'GEORGIA': 'GE',
    'GERMANY': 'DE',
    'GHANA': 'GH',
    'GIBRALTAR': 'GI',
    'GREECE': 'GR',
    'GREENLAND': 'GL',
    'GRENADA': 'GD',
    'GUADELOUPE': 'GP',
    'GUAM': 'GU',
    'GUATEMALA': 'GT',
    'GUERNSEY': 'GG',
    'GUINEA': 'GN',
    'GUINEA-BISSAU': 'GW',
    'GUYANA': 'GY',
    'HAITI': 'HT',
    'HEARD ISLAND AND MCDONALD ISLANDS': 'HM',
    'HOLY SEE (VATICAN CITY STATE)': 'VA',
    'HONDURAS': 'HN',
    'HONG KONG': 'HK',
    'HUNGARY': 'HU',
    'ICELAND': 'IS',
    'INDIA': 'IN',
    'INDONESIA': 'ID',
    'IRAN, ISLAMIC REPUBLIC OF': 'IR',
    'IRAQ': 'IQ',
    'IRELAND': 'IE',
    'ISLE OF MAN': 'IM',
    'ISRAEL': 'IL',
    'ITALY': 'IT',
    'JAMAICA': 'JM',
    'JAPAN': 'JP',
    'JERSEY': 'JE',
    'JORDAN': 'JO',
    'KAZAKHSTAN': 'KZ',
    'KENYA': 'KE',
    'KIRIBATI': 'KI',
    'KOREA, DEMOCRATIC PEOPLE\'S REPUBLIC OF': 'KP',
    'KOREA, REPUBLIC OF': 'KR',
    'KUWAIT': 'KW',
    'KYRGYZSTAN': 'KG',
    'LAO PEOPLE\'S DEMOCRATIC REPUBLIC': 'LA',
    'LATVIA': 'LV',
    'LEBANON': 'LB',
    'LESOTHO': 'LS',
    'LIBERIA': 'LR',
    'LIBYA': 'LY',
    'LIECHTENSTEIN': 'LI',
    'LITHUANIA': 'LT',
    'LUXEMBOURG': 'LU',
    'MACAO': 'MO',
    'MACEDONIA, THE FORMER YUGOSLAV REPUBLIC OF': 'MK',
    'MADAGASCAR': 'MG',
    'MALAWI': 'MW',
    'MALAYSIA': 'MY',
    'MALDIVES': 'MV',
    'MALI': 'ML',
    'MALTA': 'MT',
    'MARSHALL ISLANDS': 'MH',
    'MARTINIQUE': 'MQ',
    'MAURITANIA': 'MR',
    'MAURITIUS': 'MU',
    'MAYOTTE': 'YT',
    'MEXICO': 'MX',
    'MICRONESIA, FEDERATED STATES OF': 'FM',
    'MOLDOVA, REPUBLIC OF': 'MD',
    'MONACO': 'MC',
    'MONGOLIA': 'MN',
    'MONTENEGRO': 'ME',
    'MONTSERRAT': 'MS',
    'MOROCCO': 'MA',
    'MOZAMBIQUE': 'MZ',
    'MYANMAR': 'MM',
    'NAMIBIA': 'NA',
    'NAURU': 'NR',
    'NEPAL': 'NP',
    'NEUTRAL ZONE': 'NT',
    'NETHERLANDS': 'NL',
    'NETHERLADS ANTILLES': 'AN',
    'NEW CALEDONIA': 'NC',
    'NEW ZEALAND': 'NZ',
    'NICARAGUA': 'NI',
    'NIGER': 'NE',
    'NIGERIA': 'NG',
    'NIUE': 'NU',
    'NORFOLK ISLAND': 'NF',
    'NORTHERN MARIANA ISLANDS': 'MP',
    'NORWAY': 'NO',
    'OMAN': 'OM',
    'PAKISTAN': 'PK',
    'PALAU': 'PW',
    'PALESTINE, STATE OF': 'PS',
    'PANAMA': 'PA',
    'PAPUA NEW GUINEA': 'PG',
    'PARAGUAY': 'PY',
    'PERU': 'PE',
    'PHILIPPINES': 'PH',
    'PITCAIRN': 'PN',
    'POLAND': 'PL',
    'PORTUGAL': 'PT',
    'PUERTO RICO': 'PR',
    'QATAR': 'QA',
    'RÉUNION': 'RE',
    'ROMANIA': 'RO',
    'RUSSIAN FEDERATION': 'RU',
    'RWANDA': 'RW',
    'SAINT BARTHÉLEMY': 'BL',
    'SAINT HELENA, ASCENSION AND TRISTAN DA CUNHA': 'SH',
    'SAINT KITTS AND NEVIS': 'KN',
    'SAINT LUCIA': 'LC',
    'SAINT MARTIN (FRENCH PART)': 'MF',
    'SAINT PIERRE AND MIQUELON': 'PM',
    'SAINT VINCENT AND THE GRENADINES': 'VC',
    'SAMOA': 'WS',
    'SAN MARINO': 'SM',
    'SAO TOME AND PRINCIPE': 'ST',
    'SAUDI ARABIA': 'SA',
    'SENEGAL': 'SN',
    'SERBIA': 'RS',
    'SERBIA AND MONTENEGRO': 'AB',
    'SEYCHELLES': 'SC',
    'SIERRA LEONE': 'SL',
    'SINGAPORE': 'SG',
    'SINT MAARTEN (DUTCH PART)': 'SX',
    'SLOVAKIA': 'SK',
    'SLOVENIA': 'SI',
    'SOLOMON ISLANDS': 'SB',
    'SOMALIA': 'SO',
    'SOUTH AFRICA': 'ZA',
    'SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS': 'GS',
    'SOUTH SUDAN': 'SS',
    'SPAIN': 'ES',
    'SRI LANKA': 'LK',
    'SUDAN': 'SD',
    'SURINAME': 'SR',
    'SVALBARD AND JAN MAYEN': 'SJ',
    'SWAZILAND': 'SZ',
    'SWEDEN': 'SE',
    'SWITZERLAND': 'CH',
    'SYRIAN ARAB REPUBLIC': 'SY',
    'TAIWAN, PROVINCE OF CHINA': 'TW',
    'TAJIKISTAN': 'TJ',
    'TANZANIA, UNITED REPUBLIC OF': 'TZ',
    'THAILAND': 'TH',
    'TIMOR-LESTE': 'TL',
    'TOGO': 'TG',
    'TOKELAU': 'TK',
    'TONGA': 'TO',
    'TRINIDAD AND TOBAGO': 'TT',
    'TUNISIA': 'TN',
    'TURKEY': 'TR',
    'TURKMENISTAN': 'TM',
    'TURKS AND CAICOS ISLANDS': 'TC',
    'TUVALU': 'TV',
    'UGANDA': 'UG',
    'UKRAINE': 'UA',
    'UNITED ARAB EMIRATES': 'AE',
    'UNITED KINGDOM': 'GB',
    'UNITED STATES': 'US',
    'UNITED STATES MINOR OUTLYING ISLANDS': 'UM',
    'URUGUAY': 'UY',
    'USSR': 'SU',
    'UZBEKISTAN': 'UZ',
    'VANUATU': 'VU',
    'VENEZUELA, BOLIVARIAN REPUBLIC OF': 'VE',
    'VIET NAM': 'VN',
    'VIRGIN ISLANDS, BRITISH': 'VG',
    'VIRGIN ISLANDS, U.S.': 'VI',
    'WALLIS AND FUTUNA': 'WF',
    'WESTERN SAHARA': 'EH',
    'YEMEN': 'YE',
    'YUGOSLAVIA': 'YU',
    'ZAMBIA': 'ZM',
    'ZIMBABWE': 'ZW',
}

iso_code_to_country_name = {v: k for k, v in country_to_iso_code.items()}

countries_alternative_codes = {
    'CN': ['CHINA'],
    'FI': ['FN'],
    'FR': ['FX'],
    'GB': ['UK'],
    'TL': ['TP'],
    'CD': ['ZR']
}

countries_alternative_spellings = {
    'AU': ['SAUSTRALIA'],
    'AM': ['REPUBLIC OF ARMENIA'],
    'BA': ['BOSNIA-HERZEGOVINA', 'BOSNIA'],
    'BH': ['KINGDOM OF BAHRAIN'],
    'BJ': ['REPUBLIC OF BENIN'],
    'BO': ['BOLIVIA'],
    'BR': ['BRASIL', 'SC BRAZIL'],
    'CA': ['BC CANADA', 'BRITISH COLUMBIA'],
    'CH': ['SWITZ', 'SWITZLERAND', 'SWITZERLAND (CERN)', 'SWITERLAND'],
    'CN': ['PR CHINA'],
    'CS': ['CZECHSOLVAKIA'],
    'CZ': ['PRAGUE'],
    'DE': ['DEUTSCHLAND', 'WEST GERMANY', 'EAST GERMANY', 'BAVARIA',
           'GERMANY (DESY)'],
    'ES': ['CANARY ISLANDS', 'MADRID'],
    'FR': ['CORSICA'],
    'GR': ['CRETE'],
    'GB': ['UK', 'ENGLAND', 'ENG', 'SCOTLAND', 'WALES', 'SCOTLAND/UK',
           'NORTHERN IRELAND', 'LONDON'],
    'ID': ['BALI'],
    'IL': ['JERUSALEM'],
    'IR': ['IRAN'],
    'IT': ['ITALIE', 'ITALIA', 'NAPOLI', 'MESSINA', 'SARDINIA', 'SICILY'],
    'KR': ['SOUTH KOREA', 'SOUTH-KOREA', 'REPUBLIC OF KOREA'],
    'NL': ['THE NETHERLANDS', 'NEDERLANDS', 'NETHERLANDS'],
    'SU': ['SOVIET UNION'],
    'RS': ['REPUBLIC OF SERBIA'],
    'RU': ['RUSSIA', 'RUSSI', 'RUSSIA FEDERATION'],
    'SK': ['SLOVAK REPUBLIC'],
    'TW': ['TAIWAN'],
    'YU': ['YUGOSL', 'YU'],
    'VE': ['VENEZUELA'],
    'VN': ['VIETNAM'],
    'US': ['UNITED STATES OF AMERICA', 'UNITED STATES', 'US', 'USA'],
    'ZA': ['SAFRICA']
}


us_state_to_iso_code = {
    'ALABAMA': 'AL',
    'ALASKA': 'AK',
    'ARIZONA': 'AZ',
    'ARKANSAS': 'AR',
    'CALIFORNIA': 'CA',
    'COLORADO': 'CO',
    'CONNECTICUT': 'CT',
    'DELAWARE': 'DE',
    'DISTRICT OF COLUMBIA': 'DC',
    'FLORIDA': 'FL',
    'GEORGIA': 'GA',
    'HAWAII': 'HI',
    'IDAHO': 'ID',
    'ILLINOIS': 'IL',
    'INDIANA': 'IN',
    'IOWA': 'IA',
    'KANSAS': 'KS',
    'KENTUCKY': 'KY',
    'LOUISIANA': 'LA',
    'MAINE': 'ME',
    'MARYLAND': 'MD',
    'MASSACHUSETTS': 'MA',
    'MICHIGAN': 'MI',
    'MINNESOTA': 'MN',
    'MISSISSIPPI': 'MS',
    'MISSOURI': 'MO',
    'MONTANA': 'MT',
    'NEBRASKA': 'NE',
    'NEVADA': 'NV',
    'NEW HAMPSHIRE': 'NH',
    'NEW JERSEY': 'NJ',
    'NEW MEXICO': 'NM',
    'NEW YORK': 'NY',
    'NORTH CAROLINA': 'NC',
    'NORTH DAKOTA': 'ND',
    'OHIO': 'OH',
    'OKLAHOMA': 'OK',
    'OREGON': 'OR',
    'PENNSYLVANIA': 'PA',
    'RHODE ISLAND': 'RI',
    'SOUTH CAROLINA': 'SC',
    'SOUTH DAKOTA': 'SD',
    'TENNESSEE': 'TN',
    'TEXAS': 'TX',
    'UTAH': 'UT',
    'VERMONT': 'VT',
    'VIRGINIA': 'VA',
    'WASHINGTON': 'WA',
    'WEST VIRGINIA': 'WV',
    'WISCONSIN': 'WI',
    'WYOMING': 'WY'
}

us_states_alternative_spellings = {
    'AL': ['AL', 'ALA'],
    'AK': ['AK'],
    'AZ': ['AZ', 'ARIZ'],
    'AR': ['AR'],
    'CA': ['CA', 'CALIFORNIA', 'CALIF', 'CALF'],
    'CO': ['CO', 'COLO'],
    'CT': ['CT', 'CONNETICUT'],
    'DC': ['DC'],
    'DE': ['DE', 'DEL'],
    'FL': ['FL'],
    'GA': ['GA'],
    'HI': ['HI'],
    'ID': ['ID'],
    'IL': ['IL', 'ILL', 'ILLINOIS'],
    'IN': ['IN'],
    'IA': ['IA'],
    'KS': ['KS'],
    'KY': ['KY'],
    'LA': ['LA'],
    'ME': ['ME'],
    'MD': ['MD'],
    'MA': ['MA', 'MASS'],
    'MI': ['MI', 'MICH'],
    'MN': ['MN'],
    'MS': ['MS'],
    'MO': ['MO'],
    'MT': ['MT'],
    'NE': ['NE', 'NEV'],
    'NV': ['NV'],
    'NH': ['NH'],
    'NJ': ['NJ'],
    'NM': ['NM', 'NMEX', 'NEW MEXIKO', 'NEW MEX', 'N MEX', 'NEW MEXICO'],
    'NY': ['NY'],
    'NC': ['NC'],
    'ND': ['ND'],
    'OH': ['OH'],
    'OK': ['OK'],
    'OR': ['OR'],
    'PA': ['PA'],
    'RI': ['RI'],
    'SC': ['SC'],
    'SD': ['SD'],
    'TN': ['TN', 'TENN'],
    'TX': ['TX', 'TEX'],
    'UT': ['UT'],
    'VT': ['VT'],
    'VA': ['VA'],
    'WA': ['WA'],
    'WV': ['WV'],
    'WI': ['WI', 'WIS', 'WISC'],
    'WY': ['WY'],
}

south_korean_cities = ['SEOUL', 'DAEJON', 'DAEJEON', 'MT SORAK', 'POHANG',
                       'JEJU ISLAND', 'CHEJU ISLAND', 'GYEONGJU', 'BUSAN',
                       'DAEGU', 'GYEONGIU', 'PUSAN', 'YONGPYONG',
                       'PHOENIX PARK', 'CHEJU ISLAND']


def match_country_code(original_code):
    if isinstance(original_code, six.string_types):
        original_code = original_code.upper()
        if iso_code_to_country_name.get(original_code):
            return original_code
        else:
            for country_code, alternatives in countries_alternative_codes.items():
                for alternative in alternatives:
                    if original_code == alternative:
                        return country_code
            return None
    else:
        return None


def match_country_name_to_its_code(country_name, city=''):
    """Try to match country name with its code.

    Name of the city helps when country_name is "Korea".
    """
    if country_name:
        country_name = country_name.upper().replace('.', '').strip()

        if country_to_iso_code.get(country_name):
            return country_to_iso_code.get(country_name)
        elif country_name == 'KOREA':
            if city.upper() in south_korean_cities:
                return 'KR'
        else:
            for c_code, spellings in countries_alternative_spellings.items():
                for spelling in spellings:
                    if country_name == spelling:
                        return c_code

    return None


def match_us_state(state_string):
    """Try to match a string with one of the states in the US."""
    if state_string:
        state_string = state_string.upper().replace('.', '').strip()
        if us_state_to_iso_code.get(state_string):
            return us_state_to_iso_code.get(state_string)
        else:
            for code, state_spellings in us_states_alternative_spellings.items():
                for spelling in state_spellings:
                    if state_string == spelling:
                        return code
    return None


def parse_conference_address(address_string):
    """Parse a conference address.

    This is a pretty dummy address parser. It only extracts country
    and state (for US) and should be replaced with something better,
    like Google Geocoding.
    """

    geo_elements = address_string.split(',')
    city = geo_elements[0]
    country_name = geo_elements[-1].upper().replace('.', '').strip()
    us_state = None
    state = None
    country_code = None

    # Try to match the country
    country_code = match_country_name_to_its_code(country_name, city)

    if country_code == 'US' and len(geo_elements) > 1:
        us_state = match_us_state(geo_elements[-2].upper().strip()
                                  .replace('.', ''))

    if not country_code:
        # Sometimes the country name stores info about U.S. state
        us_state = match_us_state(country_name)

    if us_state:
        state = us_state
        country_code = 'US'

    return {
        'cities': [
            city,
        ],
        'country_code': country_code,
        'postal_code': None,
        'state': state,
    }


def parse_institution_address(address, city, state_province,
                              country, postal_code, country_code):
    """Parse an institution address."""
    address_list = force_list(address)
    state_province = match_us_state(state_province) or state_province

    postal_code = force_list(postal_code)
    country = force_list(country)
    country_code = match_country_code(country_code)

    if isinstance(postal_code, (tuple, list)):
        postal_code = ', '.join(postal_code)

    if isinstance(country, (tuple, list)):
        country = ', '.join(set(country))

    if not country_code and country:
        country_code = match_country_name_to_its_code(country)

    if not country_code and state_province and state_province in us_state_to_iso_code.values():
        country_code = 'US'

    return {
        'cities': force_list(city),
        'country_code': country_code,
        'postal_address': address_list,
        'postal_code': postal_code,
        'state': state_province,
    }



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""DoJSON common rules."""

from __future__ import absolute_import, division, print_function

import re
from datetime import datetime

from flask import current_app
from six.moves import urllib

from dojson import utils

from inspire_schemas.api import load_schema
from inspire_schemas.utils import classify_field
from inspire_utils.date import PartialDate, earliest_date
from inspire_utils.helpers import force_list, maybe_int

from ..conferences.model import conferences
from ..data.model import data
from ..experiments.model import experiments
from ..hep.model import hep, hep2marc
from ..hepnames.model import hepnames, hepnames2marc
from ..institutions.model import institutions
from ..jobs.model import jobs
from ..journals.model import journals
from ..utils import (
    force_single_element,
    get_recid_from_ref,
    get_record_ref,
)


IS_INTERNAL_UID = re.compile(r'^(inspire:uid:)?\d{5}$')
IS_ORCID = re.compile(r'^(orcid:)?\d{4}-\d{4}-\d{4}-\d{3}[0-9X]$')

WEBLINKS = {
    '00070': 'International Journal of Mathematics Server',
    '00108': 'J.Knot Theor.Ramifications',
    '00140': 'Heavy Ion Physics Server',
    '00145': 'Int. J. Math. Math. Sci. Server',
    '00154': 'J. Nonlin. Math. Phys. Server',
    '00181': 'Spektrum der Wissenschaft Server',
    '00182': 'SLAC Beam Line Server',
    '00201': 'ICFA Instrum.Bull. Server',
    '00203': 'Adv. Theor. Math. Phys. Server',
    '00211': 'KFZ Nachrichten Server',
    '00222': 'Living Reviews in Relativity Server',
    '00228': 'Nonlin. Phenom. Complex Syst. Server',
    '00238': 'Math. Phys. Anal. Geom. Server',
    '00256': 'Geometry and Topology Server',
    '00257': 'Electron.J.Diff.Eq. Server',
    '00264': 'Entropy Server',
    '00286': 'HEP Lib.Web. Server',
    '00327': 'World Sci.Lect.Notes Phys. Server',
    '00357': 'Cent. Eur. J. Phys. Server',
    '00372': 'Romanian Astronomical Journal Server',
    '00376': 'ICFA Server',
    '00411': 'Les Houches Lect. Notes Server',
    '00436': 'Int. J. Geom. Meth. Mod. Phys. Server',
    '00438': 'Phys.Part.Nucl.Lett. Server',
    '00462': 'Journal of Physics Conference Series Server',
    '00466': 'Prog.Phys. Server',
    '00480': 'SIGMA Server',
    '00484': 'Electron. J. Theor. Phys. Server',
    'AAEJA': 'Astronomy and Astrophysics Server',
    'AASRE': 'Astronomy and Astrophysics Review Server',
    'ADSABS': 'ADS Abstract Service',
    'ADSABS2': 'ADS Abstract Service',
    'AFLBD': 'Annales De La Fondation Louis De Broglie Server',
    'AHPAA': 'Ann. Henri Poincare Server',
    'AIPCONF': 'AIP Conference Server',
    'AJPIA': 'Am.J.Phys. Server',
    'ALETE': 'Astronomy Letters Server',
    'ANJOA': 'Astron.J. Server',
    'ANPYA': 'Annalen der Physik Server',
    'APCPC': 'AIP Conference Server',
    'APHYE': 'Astroparticle Physics Server',
    'APNYA': 'Ann. Phys. Server',
    'APPLA': 'Applied Physics Letter Server',
    'APPOA': 'Acta Physica Polonica Server',
    'APSSB': 'Astrophys. Space Sci. Server',
    'APSVC': 'Acta Physica Slovaca Server',
    'ARAAA': 'Annual Review of Astronomy and Astrophysics Server',
    'ARNUA': 'Annual Review of Nuclear and Particle Science',
    'ARTHF': 'Algebras and Representation Theory Server',
    'ARXERR': 'Erratum from arXiv.org server',
    'ASJOA': 'Astrophysical Journal Server',
    'ATLAS': 'ATLAS Notes from CERN server',
    'ATMP': 'ATMP Server',
    'ATPYA': 'Astrophysics Server',
    'ATROE': 'Astronomy Reports Server',
    'AUJPA': 'Austral.J.Phys. Server',
    'BABAR': 'Postscript from BABAR [BABAR Collaborators Only]',
    'BABARWEB': 'BaBar Publications Database',
    'BABARWEBP': 'BaBar Password Protected Publications Database',
    'BJPHE': 'Braz. J. Phys. Server',
    'BNL': 'Brookhaven Document Server',
    'BOOK-REVIEW': 'Review of this book',
    'BOOKCL': 'SLAC BOOKS Database',
    'BOOKREVIEW': 'Book review/article',
    'BOOKS': 'SLAC BOOKS Database',
    'BROOKHAVEN': 'Brookhaven Document Server',
    'C00-02-07': 'CHEP2000 Server',
    'C00-06-26.2': 'EPAC 2000 Server',
    'C00-09-14': 'QFTHEP00 Server',
    'C00-10-09.4': 'PCAPAC00 Server',
    'C01-01-20.1': 'Moriond Server',
    'C01-06-18.1': 'PAC01 Server',
    'C01-06-25.8': 'BeamBeam01 Server',
    'C01-07-26': 'MENU2001 Server',
    'C01-08-07': 'ICRC01 Server',
    'C01-08-20.1': 'FEL 2001 Server',
    'C01-09-06.1': 'QFTHEP01 Server',
    'C01-10-02.1': 'DESY Conference Server',
    'C02-03-18.1': 'Durham Conference Server',
    'C02-06-17.2': 'SUSY02 Server',
    'C02-09-22': 'Hadron Structure 2002 Server',
    'C02-10-15.1': 'Aguas de Lindoia 2002 Server',
    'C03-07-31': 'ICRC Server',
    'C04-04-14': 'DIS04 Server',
    'C04-04-19.4': 'ECLOUD04 Server',
    'C04-11-08.2': 'HHH-2004 Server',
    'C05-03-22': 'Phystat05 Conference Server',
    'C05-09-12': 'Phystat05 Conference Server',
    'C05-09-12.10': 'CERN Conference Server',
    'C05-10-17': 'ICATPP9 Conference Server',
    'C06-09-16.1': 'Yalta06 Conference Server',
    'C07-04-16': 'DIS2007 Conference Server',
    'C07-05-21.2': 'Blois07 Server',
    'C91-08-05': 'SLAC Summer Inst. 1991 Server',
    'C92-07-13.1': 'SLAC Summer Inst. 1992 Server',
    'C93-07-26': 'SLAC Summer Inst. 1993 Server',
    'C95-07-10.2': 'SLAC Summer Inst 1995 Server',
    'C95-10-29': 'ICALEPCS\'95 Server',
    'C97-09-04.1': 'QFTHEP\'97 Server',
    'C97-10-13': 'IWAA\'97 Server',
    'C98-06-22': 'EPAC98 Server',
    'C99-03-29': 'JACoW PAC99 Server',
    'C99-04-06': 'ANL Conference Server',
    'C99-04-12.3': 'LHC 99 JACoW Server',
    'C99-08-17': 'Conference Server',
    'C99-10-04': 'ICALEPCS99 Server',
    'CBPF': 'Rio de Janeiro, CBPF Server',
    'CDF93': 'CDF Document Server',
    'CDF94': 'CDF Document Server',
    'CDF95': 'CDF Document Server',
    'CDF96': 'CDF Document Server',
    'CDF97': 'CDF Document Server',
    'CDF98': 'CDF Document Server',
    'CDFCONF94': 'CDF Document Server',
    'CDFCONF95': 'CDF Document Server',
    'CDFCONF96': 'CDF Document Server',
    'CDFCONF97': 'CDF Document Server',
    'CDFCONF98': 'CDF Document Server',
    'CDS': 'CERN Document Server',
    'CDSWEB': 'CERN Library Record',
    'CECOA': 'CERN Courier Server',
    'CERN': 'CERN Document Server',
    'CERN-ATLAS': 'CERN ATLAS Server',
    'CERN-ATLAS-THESIS': 'ATLAS Thesis Server',
    'CERN9401': 'CERN Document Server',
    'CERN9402': 'CERN Document Server',
    'CERN9403': 'CERN Document Server',
    'CERN9404': 'CERN Document Server',
    'CERN9405': 'CERN Document Server',
    'CERN9406': 'CERN Document Server',
    'CERN9407': 'CERN Document Server',
    'CERN9408': 'CERN Document Server',
    'CERN9409': 'CERN Document Server',
    'CERN9410': 'CERN Document Server',
    'CERN9411': 'CERN Document Server',
    'CERN9412': 'CERN Document Server',
    'CERN9501': 'CERN Document Server',
    'CERN9502': 'CERN Document Server',
    'CERN9503': 'CERN Document Server',
    'CERN9504': 'CERN Document Server',
    'CERN9505': 'CERN Document Server',
    'CERN9506': 'CERN Document Server',
    'CERN9507': 'CERN Document Server',
    'CERN9508': 'CERN Document Server',
    'CERN9509': 'CERN Document Server',
    'CERN9510': 'CERN Document Server',
    'CERN9511': 'CERN Document Server',
    'CERN9512': 'CERN Document Server',
    'CERN9601': 'CERN Document Server',
    'CERN9602': 'CERN Document Server',
    'CERN9603': 'CERN Document Server',
    'CERN9604': 'CERN Document Server',
    'CERN9605': 'CERN Document Server',
    'CERN9606': 'CERN Document Server',
    'CERN9607': 'CERN Document Server',
    'CERN9608': 'CERN Document Server',
    'CERN9609': 'CERN Document Server',
    'CERN9610': 'CERN Document Server',
    'CERN9611': 'CERN Document Server',
    'CERN9612': 'CERN Document Server',
    'CERNAB': 'CERN AB Server',
    'CERNKEY': 'CERN Library Record',
    'CERNNO': 'CERN Library Record',
    'CERNREC': 'Cern Document Server',
    'CERNREP': 'CERN Server',
    'CERNSL': 'CERN SL Server',
    'CERNWEBCAST': 'CERN Web Lecture Archive',
    'CERNYEL': 'CERN Yellow Reports Server',
    'CERNYELLOW': 'CERN Yellow Reports Server',
    'CHACD': 'J. Korean Astron. Soc. Server',
    'CHAOE': 'Chaos Server',
    'CHEP97': 'Postscript from CHEP97 Server',
    'CHPHD': 'Chinese Phys. Server',
    'CITESEER': 'CiteSeer Server',
    'CJAAA': 'Chin. J. Astron. Astrophys. Server',
    'CJOPA': 'Chin.J.Phys. Server',
    'CJPHA': 'Canadian Journal of Physics Server',
    'CLNS93': 'Cornell Document Server',
    'CLNS94': 'Cornell Document Server',
    'CLNS95': 'Cornell Document Server',
    'CLNS96': 'Cornell Document Server',
    'CLNS97': 'Cornell Document Server',
    'CLNS98': 'Cornell Document Server',
    'CLNSCONF97': 'Cornell Document Server',
    'CLNSCONF98': 'Cornell Document Server',
    'CMMPE': 'Cambridge Monogr.Math.Phys. Server',
    'CMP': 'Commun. Math. Phys. Server',
    'CMPHF': 'Condensed Matter Phys. Server',
    'CNRS': 'CNRS Server',
    'COGPRINTS': 'Cogprints Server',
    'COLUMBIA-THESIS': 'Columbia U. Thesis Server',
    'CORNELL-LNS': 'Cornell U., LNS Server',
    'CPHCB': 'Comput. Phys. Commun. Server',
    'CPLEE': 'Chinese Physics Letters Server',
    'CQG': 'Class. Quant. Grav. server',
    'CQGRD': 'Class. Quantum Grav. Server',
    'CSENF': 'Comput. Sci. Engin. Server',
    'CT-THESIS': 'Caltech Thesis Server',
    'CTPHA': 'Contemp. Phys. Server',
    'CTPMD': 'Commun. Theor. Phys. Server',
    'CUP': 'Cambridge University Press Server',
    'CZYPA': 'Czech. J. Phys. Server',
    'D0': 'Postscript_Version from D0 Server',
    'DANKA': 'Doklady Physics Server',
    'DAPN': 'DAPNIA, Saclay Server',
    'DAPN-THESIS': 'DAPNIA, Saclay Thesis Server',
    'DARENET': 'DARENET Server',
    'DELPHITHESIS': 'DELPHI Thesis Server',
    'DESY': 'DESY Document Server',
    'DESY91': 'DESY Document Server',
    'DESY92': 'DESY Document Server',
    'DESY93': 'DESY Document Server',
    'DESY94': 'DESY Document Server',
    'DESY95': 'DESY Document Server',
    'DESY96': 'DESY Document Server',
    'DESYPROC': 'DESY Proceedings Server',
    'DGA': 'Diff. Geom. Appl. Document Server',
    'DMSEE': 'Turkish Journal of Math Server',
    'DOI': 'Journal Server',
    'DOPHF': 'Doklady Physics Server',
    'DSPACE': 'eCommons Digital Repository Server',
    'DUBNA': 'JINR DUBNA Preprint Server',
    'DUBNA2': 'JINR DUBNA Preprint Server',
    'DURHAM': 'HepData',
    'ECONF': 'Proceedings write-up on eConf',
    'ECONFPDF': 'pdf from eConf',
    'ECONFS': 'Slides on eConf',
    'EJP': 'Europ. J. Phys. server',
    'EJPHD': 'Eur. J. Phys. Server',
    'ELJOUR1': 'EIPL Particle Physics Server',
    'ELJOUR2': 'JHEP Server',
    'EPHJA-A': 'Euro.Phys.J.A Server',
    'EPHJA-B': 'Eur.Phys.J.B Server',
    'EPHJA-C': 'Eur.Phys.J.C Server',
    'EPHJA-D': 'Eur.Phys.J.D Server',
    'EPHJA-E': 'Eur.Phys.J.direct Server',
    'EPHJD': 'Eur.Phys.J.direct Server',
    'EPJA': 'Europ. Phys. Jour. A Server',
    'EPJB': 'Euro. Phys. Jour. B Server',
    'EPJC': 'Eur.Phys.J. C. Server',
    'EPJCD': 'Europ. Phys. Jour. direct C Server',
    'EPJD': 'Euro. Phys. Jour. D Server',
    'EPL': 'Europhys. Lett. Server',
    'EPN': 'Europhys. News Server',
    'EUCLID': 'Project Euclid Server',
    'EUPNA': 'Europhysics News Server',
    'EXASE': 'Experimental Astronomy Server',
    'FBOOKS': 'Fermilab BOOKS Database',
    'FBS': 'Few Body Sys. Server',
    'FERMILAB': 'Fermilab Library Server (fulltext available)',
    'FERMILABAPNOTE': 'Fermilab Library Server (fulltext available)',
    'FERMILABBACHELORS': 'Fermilab Library Server (fulltext available)',
    'FERMILABBEAMDOC': 'Fermilab Beam Docs Server',
    'FERMILABCONF': 'Fermilab Library Server (fulltext available)',
    'FERMILABDESIGN': 'Fermilab Library Server (fulltext available)',
    'FERMILABEN': 'Fermilab Library Server (fulltext available)',
    'FERMILABEXP': 'Fermilab Library Server (fulltext available)',
    'FERMILABFN': 'Fermilab Library Server (fulltext available)',
    'FERMILABLOI': 'Fermilab Library Server',
    'FERMILABLU': 'Fermilab Library Server (fulltext available)',
    'FERMILABMASTERS': 'Fermilab Library Server (fulltext unavailable)',
    'FERMILABMASTERSF': 'Fermilab Library Server (fulltext available)',
    'FERMILABMINERVA': 'MINERvA Document Server',
    'FERMILABMISC': 'Fermilab Library Server (fulltext available)',
    'FERMILABNAL': 'Fermilab Library Server (fulltext available)',
    'FERMILABOTHER': 'Fermilab Library Server',
    'FERMILABPROPOSAL': 'Fermilab Library Server (fulltext available)',
    'FERMILABPUB': 'Fermilab Library Server (fulltext available)',
    'FERMILABR': 'Fermilab Library Server (fulltext available)',
    'FERMILABRESEARCH': 'Fermilab Library Server (fulltext available)',
    'FERMILABSDC': 'Fermilab Library Server',
    'FERMILABTEST': 'Fermilab Library Server (fulltext available)',
    'FERMILABTHESIS': 'Fermilab Library Server (fulltext unavailable)',
    'FERMILABTHESISF': 'Fermilab Library Server (fulltext available)',
    'FERMILABTM': 'Fermilab Library Server (fulltext available)',
    'FERMILABTODAY': 'Fermilab Today Result of the Week',
    'FERMILABUPC': 'Fermilab Library Server (fulltext available)',
    'FERMILABVLHCPUB': 'Fermilab Library Server (fulltext available)',
    'FERMILABWORKBOOK': 'Fermilab Library Server',
    'FIZIKAB': 'Fizika B Server',
    'FNDPA': 'Found.Phys. Server',
    'FOCUS': 'Physical Review Focus',
    'FPLEE': 'Found. Phys. Lett. Server',
    'FPYKA': 'Fortschr. Phys. Server',
    'FZKAA-B': 'Fizika B Server',
    'FZKAAB': 'Fizika B Server',
    'GRGVA': 'Gen.Rel.Grav. Server',
    'HEPLW': 'HEP Lib.Webzine Server',
    'HEPPDF': 'PDF Server',
    'HLTPA': 'Health Physics Server',
    'HSERVER': 'HTML_Version from a server',
    'HTTP://POS.SISSA.IT/ARCHIVE/CONFERENCES/045/026/LHC07_026.PDF': 'HTTP://WWW-BD.FNAL.GOV/ICFABD/NEWSLETTER45.PDF',
    'ICTP': 'ICTP Trieste Preprint Server',
    'ICTP-LNS': 'ICTP Lecture Notes Server',
    'IEEE': 'IEEExplore Server',
    'IHEP': 'IHEP Document Server',
    'IJMPA': 'Int. J. Mod. Phys. Server',
    'IJMPB': 'Int. J. Mod. Phys. Server',
    'IJMPC': 'Int. J. Mod. Phys. Server',
    'IJMPD': 'Int. J. Mod. Phys. Server',
    'IJMPE': 'Int. J. Mod. Phys. Server',
    'IJTPB': 'Int. J. Theor. Phys. Server',
    'IMPAE': 'Int.J.Mod.Phys.A Server',
    'IMPAE-A': 'Int.J.Mod.Phys.A Server',
    'IMPAE-B': 'Int.J.Mod.Phys.B Server',
    'IMPAE-C': 'Int.J.Mod.Phys.C Server',
    'IMPAE-D': 'Int.J.Mod.Phys.D Server',
    'IMPAE-E': 'Int.J.Mod.Phys.E Server',
    'IN2P3': 'HAL in2p3 Server',
    'INDICO': 'CERN Indico Server',
    'INETA': 'Instruments and Experimental Techniques Server',
    'INTERACTIONS': 'Interactions.org article',
    'IOPLETT': 'IOP Phys.Express Lett. Server',
    'IRNPE': 'Int.Rev.Nucl.Phys. Server',
    'JACOW': 'Full-text at JACoW Server',
    'JACOWS': 'Slides on JACoW Server',
    'JAUMA': 'J.Austral.Math.Soc. Server',
    'JCAPA': 'JCAP Electronic Journal Server',
    'JCTPA': 'J. Comput. Phys. Server',
    'JDGEA': 'J. Diff. Geom. Server',
    'JETP': 'J. Exp. Theor. Phys. Server',
    'JETPL': 'J. Exp. Theor. Phys. Lett. Server',
    'JGP': 'J. Geom. Phys. Document Server',
    'JGPHE': 'Journal of Geometry and Physics Server',
    'JHEP': 'JHEP Electronic Journal Server',
    'JHEPA': 'JHEP Electronic Journal Server',
    'JHEPA-CONF': 'JHEP Conference PDF Server',
    'JHEPOA': 'JHEP Electronic Journal Server',
    'JINST': 'JINST Electronic Journal Server',
    'JKPSD': 'J. Korean Phys. Soc. Server',
    'JLAB': 'JLab Document Server',
    'JMAPA': 'J.Math.Phys. Server',
    'JMP': 'J. Math. Phys. server',
    'JPA': 'J. Phys. A server',
    'JPAGB': 'J. Phys. A Server',
    'JPCBA': 'J. Phys. C Server',
    'JPG': 'J. Phys. G server',
    'JPHGB': 'J. Phys. G Server',
    'JSTAT': 'Journal of Statistical Mechanics Server',
    'JSTOR': 'JSTOR Server',
    'JSYRE': 'J. Synchrotron Radiat. Server',
    'JTPHE': 'J.Exp.Theor.Phys. Server',
    'JTPLA': 'J.Exp.Theor.Phys.Lett. Server',
    'JUPSA': 'Journal of the Physical Society of Japan Server',
    'JVSTA': 'J. Vac. Sci. Technol. Server',
    'JVSTA-A': 'J. Vac. Sci. Technol. A server',
    'JVSTA-B': 'J. Vac. Sci. Technol. B server',
    'JVSTB': 'J. Vac. Sci. Technol. Server',
    'KEKSCAN': 'KEK scanned document',
    'LANL': 'Los Alamos Server',
    'LCLSP': 'LCLS Papers Server',
    'LCLST': 'LCLS Tech Notes Server',
    'LCNOTES': 'DESY LC Notes Server',
    'LINAC2000': 'Linac2000 Econf Server',
    'LMPHD': 'Lett. Math. Phys Server',
    'LNPHA': 'Springer Lecture Notes of Physics Server',
    'MPEJ': 'Mathematical Physics Electronic Journal Server',
    'MPHYA': 'Med.Phys. Server',
    'MPLAE-A': 'Mod.Phys.Lett.A Server',
    'MPLAE-B': 'Mod.Phys.Lett.B Server',
    'MSNET': 'Mathematical Reviews',
    'MSTCE': 'Meas. Sci. Technol. Server',
    'MTF': 'Fermilab MTF Notes Server',
    'MUNI': 'Munich U. Server',
    'NATUA': 'Nature Server',
    'NCA': 'Il Nuovo Cimento Server',
    'NCB': 'Il Nuovo Cimento Server',
    'NCC': 'Il Nuovo Cimento Server',
    'NCD': 'Il Nuovo Cimento Server',
    'NDLTD': 'Networked Digital Library Server',
    'NEWAS': 'New Astronomy Server',
    'NEWAST': 'New Astronomy Server',
    'NFKAF': 'FZK Nachr. Server',
    'NIMA': 'Nucl. Instrum. Methods A Document Server',
    'NIMB': 'Nucl. Instrum. Methods B Document Server',
    'NJOPF': 'New Journal of Physics Server',
    'NOAND': 'Nonlinear Analysis Server',
    'NOBEL': 'Nobel Foundation Server',
    'NOTES': 'Notes or further material',
    'NOVO': 'Novosibirsk, IYF Server',
    'NPA2': 'Nucl. Phys. A Document Server',
    'NPB2': 'Nucl. Phys. B Document Server',
    'NPBPS': 'Nuclear Physics Electronic',
    'NPBPS2': 'Nuclear Physics B - Proceedings Supplements',
    'NPE': 'Nuclear Physics Electronic Announcements',
    'NSENA': 'Nucl.Sci.Eng. Server',
    'NUCIA': 'Nuovo Cim. PDF Server',
    'NUCLPHYS': 'Nuclear Physics Server',
    'NUIMA-A': 'Nuclear Physics Electronic',
    'NUIMA-B': 'Nuclear Physics Electronic',
    'NUMDAM': 'NUMDAM Server',
    'NUMI-PUBLIC': 'NuMI Server',
    'NUMI-RESTRICTED': 'NuMI Restricted Server',
    'NUPHA-A': 'Nuclear Physics Electronic',
    'NUPHA-B': 'Nuclear Physics Electronic',
    'NUPHZ': 'Nuclear Physics Electronic',
    'NUPHZ-TOC': 'Nuclear Physics Electronic',
    'NWSCA': 'New Scientist Server',
    'OLAL': 'Orsay, LAL Server',
    'OSTI': 'OSTI Information Bridge Server',
    'OUP': 'Oxford University Press Server',
    'PANUE': 'Phys.At.Nucl. Server',
    'PANUEO': 'Phys.At.Nucl. Server',
    'PARTICLEZ': 'particlez.org Server',
    'PDG': 'Particle Data Group (PDG) Server',
    'PDG-RPP': 'Review of Particle Properties full record',
    'PDG2002PDF': 'PDF from PDG site',
    'PDG2002PS': 'Postscript from PDG Site',
    'PDG2004PDF': 'PDF from PDG site',
    'PDG2004PS': 'PS from PDG site',
    'PDG98': 'PDG-RPP Server',
    'PDG98R': 'PDG Server',
    'PDGJOURNAL': 'Review of Particle Properties full record',
    'PDGLIV': 'pdgLive (measurements quoted by PDG)',
    'PHLTA-A': 'Phys. Lett. A Server',
    'PHLTA-B': 'Nuclear Physics Electronic',
    'PHMBA': 'Phys. Med. Biol. Server',
    'PHPEF': 'Physics in Perspective Server',
    'PHRVA': 'Phys.Rev. Server',
    'PHRVA-A': 'Phys. Rev. A Server',
    'PHRVA-B': 'Phys. Rev. B Server',
    'PHRVA-C': 'Phys. Rev. C Server',
    'PHRVA-D': 'Phys. Rev. D Server',
    'PHRVA-E': 'Phys. Rev. E Server',
    'PHRVA-FOCUS': 'Physical Review Focus',
    'PHSTB': 'Physica Scripta Server',
    'PHTOA': 'Physics Today Server',
    'PHUSE': 'Physics uspekhii Server',
    'PHUZA': 'Physik in unserer Zeit Server',
    'PHWOE': 'Physics World Server',
    'PHYSA': 'Physica A Document Server',
    'PHYSA-A': 'Physica A Server',
    'PHYSA-D': 'Physica D Server',
    'PHYSD': 'Physica A Document Server',
    'PHYSICSWEB': 'Physicsweb.org article',
    'PHYSICSWORLD': 'physicsworld.com article',
    'PHYSORG': 'PhysOrg.com article',
    'PHYSREV': 'Physical Review Server',
    'PNASA': 'Proc.Nat.Acad.Sci. Server',
    'POS': 'Proceedings of Science Server',
    'PPN': 'Phys. Part. Nucl. Server',
    'PPNP': 'Prog. Part. Nucl. Phys. Document Server',
    'PPNPD': 'Prog. Part. Nucl. Phys. server',
    'PPNUE': 'Phys.Part.Nucl. Server',
    'PPNUE-S': 'Phys. Part. Nucl. Server',
    'PPNUE1': 'Phys. Part. Nucl. (AIP) Server',
    'PPNUES': 'Phys. Part. Nucl. Suppl. Server',
    'PR': 'Physics Reports Document Server',
    'PRAMC': 'Pramana Server',
    'PRAMCARC': 'Pramana Archive Server',
    'PRC': 'Phys. Rev. C Server',
    'PRD': 'Phys. Rev. D Server',
    'PRDOLA': 'Phys. Rev. Online Archive',
    'PRE': 'Phys. Rev. E Server',
    'PRL': 'Phys. Rev. Lett. Server',
    'PRLTA': 'Phys. Rev. Lett. Server',
    'PRPLC': 'Physics Reports Server',
    'PRSLA': 'Proceedings of the Royal Society Server',
    'PRSTA': 'Phys.Rev.ST Accel.Beams Server',
    'PTPKA': 'Prog.Theor.Phys. Server',
    'PTRSA': 'Phil.Trans.Roy.Soc.Lond. Server',
    'PUBMED': 'PUBMED Server',
    'RCCHB': 'La Recherche Server',
    'RESOF': 'Resonance Journal of Science Education Server',
    'RINGBERG': 'Ringberg Conference Server',
    'RJMPE': 'Russ. J. Math. Phys. Server',
    'RMEAE': 'Radiation Measurements Server',
    'RMHPB': 'Reports on Mathematical Physics Server',
    'RMP': 'Rev. Mod. Phys. Server',
    'RMPHA': 'Rev. Mod. Phys. Server',
    'RMPHE': 'Reviews in Mathematical Physics Server',
    'RMXFA': 'Rev. Mex. Fis. Server',
    'RORPE': 'Rom. Rep. Phys. Server',
    'RPDOD': 'Radiation Protection Dosimetry Server',
    'RPPHA': 'Rep. Prog. Phys. Server',
    'RRPQA': 'Rom. J. Phys. Server',
    'RSI': 'Rev.Sci.Instrum. Server',
    'RSINA': 'Rev. Sci. Instrum. Server',
    'SACLAY': 'Saclay Document Server',
    'SALAM': 'ICTP Preprint Archive',
    'SCAMA': 'Scientific American Server',
    'SCIDIR': 'Science Direct',
    'SCIEA': 'Science Server',
    'SCIENCEBLOG': 'Science Blog article',
    'SCIENCEDAILY': 'Science Daily Article',
    'SERVER': 'Electronic Version from a server',
    'SERVER2': 'Electronic Version from another server',
    'SLAC': 'SLAC Document Server',
    'SLACBOOK': 'SLAC Book Catalog Record',
    'SLACPUB': 'SLAC Document Server',
    'SLACREPT': 'SLAC Document Server',
    'SLACTN': 'SLAC Document Server',
    'SLACTODAY': 'SLAC Today article',
    'SOPJA': 'Russ.Phys.J. Server',
    'SPACE': 'SPACE.com Article',
    'SPIE': 'SPIE Server',
    'SPRINGER': 'Springer Books Server',
    'SPTPA': 'Tech. Phys. Server',
    'SSC': 'HEPNET Document Server',
    'SSI91': 'SLAC Document Server',
    'SSI92': 'SLAC Document Server',
    'SSI93': 'SLAC Document Server',
    'SSI94': 'SLAC Document Server',
    'SSI96': 'SLAC Document Server',
    'SSI97': 'SLAC Document Server',
    'SSI98': 'SLAC Document Server',
    'TECHPHYS': 'Tech. Phys. Server',
    'TECHPHYSL': 'Tech. Phys. Lett. Server',
    'TJPHE': 'Turkish Journal of Physics Server',
    'TMFZA': 'Teor.Mat.Fiz. Server',
    'TMPHA': 'Theor.Math.Phys. Server',
    'TORONTO-A': 'Toronto U. Astron. Thesis Server',
    'TP': 'Tech. Phys. Server',
    'UCOL': 'University Coll. London Server',
    'UMI': 'UMI Thesis Server',
    'VAC': 'Vacuum Document Server',
    'VACUA': 'Vacuum Server',
    'VIDEO': 'Watch the video of the talk',
    'WORLDSCI': 'World Scientific Books Server',
    'ZBLATT': 'Zentralblatt MATH Server',
    'ZFDOF': 'Journal of Physical Studies Server',
    'ZNTFA': 'Z. Naturforsch. Server',
    'ZPA': 'Z. Phys. A Server',
    'ZPC': 'Z. Phys. C Server',
}


def control_number(endpoint):
    """Populate the ``control_number`` key.

    Also populates the ``self`` key through side effects.
    """
    def _control_number(self, key, value):
        self['self'] = get_record_ref(int(value), endpoint)
        return int(value)

    return _control_number


conferences.over('control_number', '^001')(control_number('conferences'))
data.over('control_number', '^001')(control_number('data'))
experiments.over('control_number', '^001')(control_number('experiments'))
hep.over('control_number', '^001')(control_number('literature'))
hepnames.over('control_number', '^001')(control_number('authors'))
institutions.over('control_number', '^001')(control_number('institutions'))
jobs.over('control_number', '^001')(control_number('jobs'))
journals.over('control_number', '^001')(control_number('journals'))


@hep2marc.over('001', '^control_number$')
@hepnames2marc.over('001', '^control_number$')
def control_number2marc(self, key, value):
    return value


@conferences.over('legacy_version', '^005')
@data.over('legacy_version', '^005')
@experiments.over('legacy_version', '^005')
@hep.over('legacy_version', '^005')
@hepnames.over('legacy_version', '^005')
@institutions.over('legacy_version', '^005')
@jobs.over('legacy_version', '^005')
@journals.over('legacy_version', '^005')
def legacy_version(self, key, value):
    return value


@hep2marc.over('005', '^legacy_version$')
@hepnames2marc.over('005', '^legacy_version$')
def legacy_version2marc(self, key, value):
    return value


@hep.over('acquisition_source', '^541..')
@hepnames.over('acquisition_source', '^541..')
def acquisition_source(self, key, value):
    """Populate the ``acquisition_source`` key."""
    def _get_datetime(value):
        d_value = force_single_element(value.get('d', ''))
        if d_value:
            try:
                date = PartialDate.loads(d_value)
            except ValueError:
                return d_value
            else:
                datetime_ = datetime(year=date.year, month=date.month, day=date.day)
                return datetime_.isoformat()

    internal_uid, orcid, source = None, None, None

    a_values = force_list(value.get('a'))
    for a_value in a_values:
        if IS_INTERNAL_UID.match(a_value):
            if a_value.startswith('inspire:uid:'):
                internal_uid = int(a_value[12:])
            else:
                internal_uid = int(a_value)
        elif IS_ORCID.match(a_value):
            if a_value.startswith('orcid:'):
                orcid = a_value[6:]
            else:
                orcid = a_value
        else:
            source = a_value

    c_value = force_single_element(value.get('c', ''))
    normalized_c_value = c_value.lower()

    if normalized_c_value == 'batchupload':
        method = 'batchuploader'
    elif normalized_c_value == 'submission':
        method = 'submitter'
    else:
        method = normalized_c_value

    return {
        'datetime': _get_datetime(value),
        'email': value.get('b'),
        'internal_uid': internal_uid,
        'method': method,
        'orcid': orcid,
        'source': source,
        'submission_number': value.get('e'),
    }


@hep2marc.over('541', '^acquisition_source$')
@hepnames2marc.over('541', '^acquisition_source$')
def acquisition_source2marc(self, key, value):
    orcid = value.get('orcid')
    source = value.get('source')

    a_value = 'orcid:' + orcid if orcid else source

    method = value.get('method')

    if method == 'batchuploader':
        c_value = 'batchupload'
    elif method == 'submitter':
        c_value = 'submission'
    else:
        c_value = method

    return {
        'a': a_value,
        'b': value.get('email'),
        'c': c_value,
        'd': value.get('datetime'),
        'e': value.get('submission_number'),
    }


@conferences.over('public_notes', '^500..')
@experiments.over('public_notes', '^500..')
@hepnames.over('public_notes', '^500..')
@institutions.over('public_notes', '^500..')
@jobs.over('public_notes', '^500..')
@journals.over('public_notes', '^500..')
@utils.flatten
@utils.for_each_value
def public_notes_500(self, key, value):
    """Populate the ``public_notes`` key."""
    return [
        {
            'source': value.get('9'),
            'value': public_note,
        } for public_note in force_list(value.get('a'))
    ]


@hep2marc.over('500', '^public_notes$')
@hepnames2marc.over('500', '^public_notes$')
@utils.for_each_value
def public_notes2marc(self, key, value):
    return {
        '9': value.get('source'),
        'a': value.get('value'),
    }


@conferences.over('_private_notes', '^595..')
@experiments.over('_private_notes', '^595..')
@hepnames.over('_private_notes', '^595..')
@institutions.over('_private_notes', '^595..')
@jobs.over('_private_notes', '^595..')
@journals.over('_private_notes', '^595..')
@utils.flatten
@utils.for_each_value
def _private_notes_595(self, key, value):
    """Populate the ``_private_notes`` key."""
    return [
        {
            'source': value.get('9'),
            'value': _private_note,
        } for _private_note in force_list(value.get('a'))
    ]


@hep2marc.over('595', '^_private_notes$')
@hepnames2marc.over('595', '^_private_notes$')
@utils.for_each_value
def _private_notes2marc(self, key, value):
    return {
        '9': value.get('source'),
        'a': value.get('value'),
    }


@conferences.over('inspire_categories', '^65017')
@experiments.over('inspire_categories', '^65017')
@hep.over('inspire_categories', '^65017')
@institutions.over('inspire_categories', '^65017')
@jobs.over('inspire_categories', '^65017')
@journals.over('inspire_categories', '^65017')
def inspire_categories(self, key, value):
    schema = load_schema('elements/inspire_field')
    valid_sources = schema['properties']['source']['enum']

    inspire_categories = self.get('inspire_categories', [])

    scheme = force_single_element(value.get('2'))
    if scheme == 'arXiv':          # XXX: we skip arXiv categories here because
        return inspire_categories  # we're going to add them later in a filter.

    source = force_single_element(value.get('9', '')).lower()
    if source not in valid_sources:
        if source == 'automatically added based on dcc, ppf, dk':
            source = 'curator'
        elif source == 'submitter':
            source = 'user'
        else:
            source = None

    terms = force_list(value.get('a'))
    for _term in terms:
        term = classify_field(_term)
        if term:
            inspire_categories.append({
                'term': term,
                'source': source,
            })

    return inspire_categories


@hep2marc.over('65017', '^inspire_categories$')
@utils.for_each_value
def inspire_categories2marc(self, key, value):
    return {
        '2': 'INSPIRE',
        '9': value.get('source'),
        'a': value.get('term'),
    }


@conferences.over('_private_notes', '^667..')
@experiments.over('_private_notes', '^667..')
@hep.over('_private_notes', '^667..')
@institutions.over('_private_notes', '^667..')
@jobs.over('_private_notes', '^667..')
@utils.for_each_value
def _private_notes_667(self, key, value):
    return {
        'source': value.get('9'),
        'value': value.get('a'),
    }


@conferences.over('public_notes', '^680..')
@experiments.over('public_notes', '^680..')
@institutions.over('public_notes', '^680..')
@jobs.over('public_notes', '^680..')
@journals.over('public_notes', '^680..')
@utils.for_each_value
def public_notes_680(self, key, value):
    return {
        'source': value.get('9'),
        'value': value.get('i'),
    }


@conferences.over('urls', '^8564.')
@experiments.over('urls', '^8564.')
@hep.over('urls', '^8564.')
@institutions.over('urls', '^8564.')
@jobs.over('urls', '^8564.')
@journals.over('urls', '^8564.')
def urls(self, key, value):
    def _is_internal_url(url):
        base = urllib.parse.urlparse(current_app.config['LEGACY_BASE_URL'])
        base_netloc = base.netloc or base.path
        parsed_url = urllib.parse.urlparse(url)
        url_netloc = parsed_url.netloc or parsed_url.path

        return base_netloc == url_netloc

    urls = self.get('urls', [])

    description = force_single_element(value.get('y'))
    description = WEBLINKS.get(description, description)
    for url in force_list(value.get('u')):
        if not _is_internal_url(url):
            urls.append({
                'description': description,
                'value': url,
            })

    return urls


@hep2marc.over('8564', '^urls$')
@hepnames2marc.over('8564', '^urls$')
@utils.for_each_value
def urls2marc(self, key, value):
    return {
        'u': value.get('value'),
        'y': value.get('description'),
    }


@conferences.over('legacy_creation_date', '^961..')
@experiments.over('legacy_creation_date', '^961..')
@hep.over('legacy_creation_date', '^961..')
@hepnames.over('legacy_creation_date', '^961..')
@institutions.over('legacy_creation_date', '^961..')
@jobs.over('legacy_creation_date', '^961..')
@journals.over('legacy_creation_date', '^961..')
def legacy_creation_date(self, key, value):
    if 'legacy_creation_date' in self and self['legacy_creation_date']:
        return self['legacy_creation_date']

    x_values = force_list(value.get('x'))
    if x_values:
        return earliest_date(x_values)


@hep2marc.over('961', '^legacy_creation_date$')
@hepnames2marc.over('961', '^legacy_creation_date$')
def legacy_creation_date2marc(self, key, value):
    return {'x': value}


def external_system_identifiers(endpoint):
    """Populate the ``external_system_identifiers`` key.

    Also populates the ``new_record`` key through side effects.
    """
    @utils.flatten
    @utils.for_each_value
    def _external_system_identifiers(self, key, value):
        new_recid = maybe_int(value.get('d'))
        if new_recid:
            self['new_record'] = get_record_ref(new_recid, endpoint)

        return [
            {
                'schema': 'SPIRES',
                'value': ext_sys_id,
            } for ext_sys_id in force_list(value.get('a'))
        ]

    return _external_system_identifiers


conferences.over('external_system_identifiers', '^970..')(external_system_identifiers('conferences'))
experiments.over('external_system_identifiers', '^970..')(external_system_identifiers('experiments'))
hep.over('external_system_identifiers', '^970..')(external_system_identifiers('literature'))
institutions.over('external_system_identifiers', '^970..')(external_system_identifiers('institutions'))
jobs.over('external_system_identifiers', '^970..')(external_system_identifiers('jobs'))
journals.over('external_system_identifiers', '^970..')(external_system_identifiers('journals'))


@hep2marc.over('970', '^new_record$')
@hepnames2marc.over('970', '^new_record$')
def new_record2marc(self, key, value):
    return {'d': get_recid_from_ref(value)}


@conferences.over('deleted', '^980..')
@data.over('deleted', '^980..')
def deleted(self, key, value):
    return value.get('c', '').upper() == 'DELETED'


def deleted_records(endpoint):
    """Populate the ``deleted_records`` key."""
    @utils.for_each_value
    def _deleted_records(self, key, value):
        deleted_recid = maybe_int(value.get('a'))
        if deleted_recid:
            return get_record_ref(deleted_recid, endpoint)

    return _deleted_records


conferences.over('deleted_records', '^981..')(deleted_records('conferences'))
data.over('deleted_records', '^981..')(deleted_records('data'))
experiments.over('deleted_records', '^981..')(deleted_records('experiments'))
hep.over('deleted_records', '^981..')(deleted_records('literature'))
hepnames.over('deleted_records', '^981..')(deleted_records('authors'))
institutions.over('deleted_records', '^981..')(deleted_records('institutions'))
jobs.over('deleted_records', '^981..')(deleted_records('jobs'))
journals.over('deleted_records', '^981..')(deleted_records('journals'))


@hep2marc.over('981', 'deleted_records')
@hepnames2marc.over('981', 'deleted_records')
@utils.for_each_value
def deleted_records2marc(self, key, value):
    return {'a': get_recid_from_ref(value)}



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""DoJSON rules for CDS to INSPIRE HEP MARC."""

from __future__ import absolute_import, division, print_function

import os
import re
from itertools import chain

import pycountry
import rfc3987
import six

from idutils import is_arxiv
from dojson import utils

from six.moves import urllib

from inspire_utils.helpers import force_list
from inspire_utils.name import normalize_name

from .model import cds2hep_marc
from ..utils import force_single_element, quote_url

CATEGORIES = {
    'General Relativity and Cosmology': 'Gravitation and Cosmology',
    'General Theoretical Physics': 'General Physics',
    'Detectors and Experimental Techniques': 'Instrumentation',
    'Engineering': 'Instrumentation',
    'Accelerators and Storage Rings': 'Accelerators',
    'Computing and Computers': 'Computing',
    'Mathematical Physics and Mathematics': 'Math and Math Physics',
    'Astrophysics and Astronomy': 'Astrophysics',
    'Condensed Matter': 'General Physics',
    'General Theoretical Physics': 'General Physics',
    'Physics in General': 'General Physics',
    'Other': 'Other',
    'Chemical Physics and Chemistry': 'Other',
    'Information Transfer and Management': 'Other',
    'Commerce, Economics, Social Science': 'Other',
    'Biography, Geography, History': 'Other',
    'Science in General': 'Other',
    'Particle Physics - Experiment': 'Experiment-HEP',
    'Particle Physics - Phenomenology': 'Phenomenology-HEP',
    'Particle Physics - Theory': 'Theory-HEP',
    'Particle Physics - Lattice': 'Lattice',
    'Nuclear Physics - Experiment': 'Experiment-Nucl',
    'Nuclear Physics - Theory': 'Theory-Nucl',
}


EXPERIMENTS = {
    ('CERN LEP', 'L3'): 'CERN-LEP-L3',
    ('CERN LEP', 'OPAL'): 'CERN-LEP-OPAL',
    ('CERN LHC', 'ALICE'): 'CERN-LHC-ALICE',
    ('CERN LHC', 'ATLAS'): 'CERN-LHC-ATLAS',
    ('CERN LHC', 'CMS'): 'CERN-LHC-CMS',
    ('CERN LHC', 'LHCb'): 'CERN-LHC-LHCb',
    ('CERN LHC', 'LHCf'): 'CERN-LHC-LHCf',
    ('CERN LHC', 'MoEDAL'): 'CERN-LHC-MoEDAL',
    ('CERN LHC', 'TOTEM'): 'CERN-LHC-TOTEM',
    ('CERN PS', 'nTOF'): 'CERN-nTOF',
    ('CERN SPS', 'ICARUS CNGS2'): 'ICARUS',
    ('CERN SPS', 'OPERA CNGS1'): 'OPERA',
    ('DESY HERA', 'ZEUS'): 'DESY-HERA-ZEUS',
}


RE_IDS = re.compile(r'\((?P<schema>.*?)\)(?P<id>.*)')


def add_source(field, source='CDS'):
    if not field.get('9'):
        field['9'] = source


def vanilla_dict(god):
    return {k: v for (k, v) in six.iteritems(god) if k != '__order__'}


def ignore_not_applicable(text):
    return text if text.lower() != 'not applicable' else None


def escape_url(url):
    try:
        rfc3987.parse(url, rule="URI")
        return url
    except ValueError:
        if url.lower().startswith('https://'):
            scheme = 'https://'
        elif url.lower().startswith('http://'):
            scheme = 'http://'
        else:
            scheme = ''

        url = quote_url(url[len(scheme):])
        return scheme + url


@cds2hep_marc.over('0247_', '^0247.')
@utils.for_each_value
def persistent_identifiers(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('035__', '^035..')
@utils.for_each_value
def external_sytem_identifiers(self, key, value):
    ignored = {'cercer', 'inspire', 'xx', 'cern annual report', 'cmscms', 'wai01', 'spires'}
    if any(val.lower() in ignored for val in chain(force_list(value.get('9')), force_list(value.get('a')))):
        return
    if any(val.lower().endswith('cercer') for val in force_list(value.get('a'))):
        return

    return vanilla_dict(value)


@cds2hep_marc.over('037__', '^037..', '^088..')
def secondary_report_numbers(self, key, value):
    """Populate the ``037`` MARC field.

    Also populates the ``500``, ``595`` and ``980`` MARC field through side effects.
    """
    preliminary_results_prefixes = ['ATLAS-CONF-', 'CMS-PAS-', 'CMS-DP-', 'LHCB-CONF-']
    note_prefixes = ['ALICE-INT-', 'ATL-', 'ATLAS-CONF-', 'CMS-DP-', 'CMS-PAS-', 'LHCB-CONF-', 'LHCB-PUB-']

    result_037 = self.get('037__', [])
    result_500 = self.get('500__', [])
    result_595 = self.get('595__', [])
    result_980 = self.get('980__', [])

    report = force_single_element(value.get('a', ''))
    hidden_report = force_single_element(value.get('9') or value.get('z', ''))
    source = 'CDS' if not is_arxiv(report) else 'arXiv'

    if any(report.upper().startswith(prefix) for prefix in note_prefixes):
        result_980.append({'a': 'NOTE'})

    if any(report.upper().startswith(prefix) for prefix in preliminary_results_prefixes):
        result_500.append({'9': 'CDS', 'a': 'Preliminary results'})

    is_barcode = hidden_report.startswith('P0') or hidden_report.startswith('CM-P0')
    if not report.startswith('SIS-') and not is_barcode:
        result_037.append({
            '9': source,
            'a': report,
            'c': value.get('c'),
            'z': hidden_report if source == 'CDS' else None,
        })

    self['500__'] = result_500
    self['595__'] = result_595
    self['980__'] = result_980
    return result_037


@cds2hep_marc.over('041__', '^041..')
@utils.flatten
@utils.for_each_value
def languages(self, key, value):
    languages = []
    values = force_list(value.get('a'))

    for language in values:
        alpha_3 = language.strip().lower()
        try:
            languages.append({'a': pycountry.languages.get(alpha_3=alpha_3).name})
        except KeyError:
            try:
                languages.append({'a': pycountry.languages.get(bibliographic=alpha_3).name})
            except KeyError:
                pass

    return languages


def _converted_author(value):
    def _get_ids_from_0(subfield):
        """Transform IDs from CDS into INSPIRE-style IDs."""
        ids = {}
        ids_i = []
        ids_j = []
        segments = subfield.split('|')

        for segment in segments:
            match = RE_IDS.match(segment)
            if match:
                ids[match.group('schema').upper()] = match.group('id')

        for schema, id_ in ids.items():
            if schema == 'INSPIRE':
                ids_i.append(id_)
            elif schema == 'SZGECERN':
                ids_j.append(u'CCID-{}'.format(id_))
            elif schema == 'CDS':
                continue
            else:
                ids_j.append(id_)

        return ids_i, ids_j

    value = vanilla_dict(value)

    if 'beard' in value.get('9', '').lower():
        value.pop('0', None)
        return value

    subfields_i = force_list(value.get('i'))
    subfields_j = force_list(value.get('j'))

    for id_ in force_list(value.pop('0', None)):
        ids_i, ids_j = _get_ids_from_0(id_)
        subfields_i.extend(ids_i)
        subfields_j.extend(ids_j)

    value['a'] = normalize_name(value['a'])
    value['i'] = subfields_i
    value['j'] = subfields_j

    return value


@cds2hep_marc.over('100__', '^100..')
@utils.for_each_value
def first_author(self, key, value):
    return _converted_author(value)


@cds2hep_marc.over('700__', '^700..')
def nonfirst_authors(self, key, value):
    """Populate ``700`` MARC field.

    Also populates the ``701`` MARC field through side-effects.
    """
    field_700 = self.get('700__', [])
    field_701 = self.get('701__', [])

    is_supervisor = any(el.lower().startswith('dir') for el in force_list(value.get('e', '')))
    if is_supervisor:
        field_701.append(_converted_author(value))
    else:
        field_700.append(_converted_author(value))

    self['701__'] = field_701
    return field_700


@cds2hep_marc.over('110__', '^110..')
@utils.for_each_value
def corporate_authors(self, key, value):
    if 'a' in value:
        return vanilla_dict(value)


@cds2hep_marc.over('242__', '^242..')
def translated_title(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('245__', '^245..')
def title(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('246__', '^246..')
@utils.for_each_value
def other_titles(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('260__', '^260..')
def imprint(self, key, value):
    return vanilla_dict(value)


@cds2hep_marc.over('300__', '^300..')
def number_of_pages(self, key, value):
    match = re.match(r'(?P<pages>\d+)', value.get('a', ''))
    if match:
        return {'a': match.group('pages')}


@cds2hep_marc.over('502__', '^502..')
def thesis_info(self, key, value):
    return {
        'b': value.get('a'),
        'c': value.get('b'),
        'd': value.get('c'),
    }


@cds2hep_marc.over('500__', '^500..')
@utils.for_each_value
def public_notes(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('520__', '^520..')
@utils.for_each_value
def abstracts(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('65017', '^65017')
@utils.for_each_value
def categories(self, key, value):
    schema = value.get('2', '')
    if schema.lower() == 'szgecern':
        result = {
            '2': 'INSPIRE',
            # XXX: will fail validation and be logged if invalid category
            'a': CATEGORIES.get(value.get('a'), value.get('a'))
        }
    else:
        result = vanilla_dict(value)

    add_source(result)
    return result


@cds2hep_marc.over('6531_', '^6531.')
@utils.for_each_value
def keywords(self, key, value):
    value = vanilla_dict(value)
    add_source(value)
    return value


@cds2hep_marc.over('693__', '^693..')
@utils.for_each_value
def accelerator_experiments(self, key, value):
    accelerator = ignore_not_applicable(value.get('a'))
    experiment = ignore_not_applicable(value.get('e'))
    experiment = EXPERIMENTS.get((accelerator, experiment), experiment)
    return {
        'a': accelerator,
        'e': experiment,
    }


@cds2hep_marc.over('65017', '^695..')
@utils.for_each_value
def arxiv_categories(self, key, value):
    is_arxiv = value.get('9', '').lower() == 'lanl eds'
    if is_arxiv:
        return {
            '2': 'arXiv',
            'a': value.get('a'),
        }


@cds2hep_marc.over('710__', '^710..')
@utils.for_each_value
def collaborations(self, key, value):
    if 'g' in value:
        return vanilla_dict(value)


@cds2hep_marc.over('773__', '^773..')
@utils.for_each_value
def publication_info(self, key, value):
    return vanilla_dict(value)


@cds2hep_marc.over('8564_', '^8564.')
def urls(self, key, value):
    """Populate the ``8564`` MARC field.

    Also populate the ``FFT`` field through side effects.
    """
    def _is_preprint(value):
        return value.get('y', '').lower() == 'preprint'

    def _is_fulltext(value):
        return value['u'].endswith('.pdf') and value['u'].startswith('http://cds.cern.ch')

    def _is_local_copy(value):
        return 'local copy' in value.get('y', '')

    def _is_ignored_domain(value):
        ignored_domains = ['http://cdsweb.cern.ch', 'http://cms.cern.ch',
                           'http://cmsdoc.cern.ch', 'http://documents.cern.ch',
                           'http://preprints.cern.ch', 'http://cds.cern.ch',
                           'http://arxiv.org']
        return any(value['u'].startswith(domain) for domain in ignored_domains)

    field_8564 = self.get('8564_', [])
    field_FFT = self.get('FFT__', [])

    if 'u' not in value:
        return field_8564

    url = escape_url(value['u'])

    if _is_fulltext(value) and not _is_preprint(value):
        if _is_local_copy(value):
            description = value.get('y', '').replace('local copy', 'on CERN Document Server')
            field_8564.append({
                'u': url,
                'y': description,
            })
        else:
            _, file_name = os.path.split(urllib.parse.urlparse(value['u']).path)
            _, extension = os.path.splitext(file_name)
            field_FFT.append({
                't': 'CDS',
                'a': url,
                'd': value.get('y', ''),
                'n': file_name,
                'f': extension,
            })
    elif not _is_ignored_domain(value):
        field_8564.append({
            'u': url,
            'y': value.get('y'),
        })

    self['FFT__'] = field_FFT
    return field_8564


@cds2hep_marc.over('980__', '^980..')
@utils.for_each_value
def collections(self, key, value):
    allowed_collections = {'note', 'thesis', 'conferencepaper'}
    collection = value.get('a', '').lower()

    if collection not in allowed_collections:
        return
    return vanilla_dict(value)


@cds2hep_marc.over('980__', '^962..')
@utils.for_each_value
def conference_paper(self, key, value):
    is_conference_paper = value.get('n', '')[-2:].isdigit()
    if is_conference_paper:
        return {'a': 'ConferencePaper'}



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""DoJSON rules for HEPNames."""

from __future__ import absolute_import, division, print_function

import re

from dojson import utils

from inspire_schemas.api import load_schema
from inspire_schemas.utils import (
    normalize_arxiv_category,
    valid_arxiv_categories,
)
from inspire_utils.date import normalize_date
from inspire_utils.helpers import force_list, maybe_int
from inspire_utils.name import normalize_name

from .model import hepnames, hepnames2marc
from ..utils import (
    force_single_element,
    get_record_ref,
    get_recid_from_ref,
    normalize_rank,
    quote_url,
    unquote_url
)


AWARD_YEAR = re.compile(r'\(?(?P<year>\d{4})\)?')
INSPIRE_BAI = re.compile(r'(\w+\.)+\d+')
LOOKS_LIKE_CERN = re.compile(r'^\d+$|^CER[MN]?-|^CNER-|^CVERN-', re.I)
NON_DIGIT = re.compile(r'[^\d]+')
LINKEDIN_URL = re.compile(r'https?://(\w+\.)?linkedin\.com/in/(?P<page>[\w%-]+)', re.UNICODE)
TWITTER_URL = re.compile(r'https?://(www\.)?twitter\.com/(?P<handle>\w+)')
WIKIPEDIA_URL = re.compile(r'https?://(?P<lang>\w+)\.wikipedia\.org/wiki/(?P<page>.*)')


@hepnames.over('ids', '^035..')
@utils.for_each_value
def ids(self, key, value):
    def _get_schema(value):
        IDS_MAP = {
            'ARXIV': 'ARXIV',
            'BAI': 'INSPIRE BAI',
            'CERN': 'CERN',
            'DESY': 'DESY',
            'GOOGLESCHOLAR': 'GOOGLESCHOLAR',
            'INSPIRE': 'INSPIRE ID',
            'KAKEN': 'KAKEN',
            'ORCID': 'ORCID',
            'RESEARCHID': 'RESEARCHERID',
            'RESEARCHERID': 'RESEARCHERID',
            'SLAC': 'SLAC',
            'SCOPUS': 'SCOPUS',
            'TWITTER': 'TWITTER',
            'VIAF': 'VIAF',
            'WIKIPEDIA': 'WIKIPEDIA',
        }

        return IDS_MAP.get(value.get('9', '').upper())

    def _guess_schema_from_value(a_value):
        if a_value is None:
            return None

        if INSPIRE_BAI.match(a_value):
            return 'INSPIRE BAI'

    def _try_to_correct_value(schema, a_value):
        if a_value is None:
            return a_value

        if schema == 'CERN' and LOOKS_LIKE_CERN.match(a_value):
            return 'CERN-' + NON_DIGIT.sub('', a_value)
        elif schema == 'KAKEN':
            return 'KAKEN-' + a_value
        else:
            return a_value

    a_value = force_single_element(value.get('a'))

    schema = _get_schema(value)
    if schema is None:
        schema = _guess_schema_from_value(a_value)

    a_value = _try_to_correct_value(schema, a_value)

    if schema and a_value:
        return {
            'schema': schema,
            'value': a_value,
        }


@hepnames2marc.over('035', '^ids$')
@utils.for_each_value
def ids2marc(self, key, value):
    """Populate the ``035`` MARC field.

    Also populates the ``8564`` and ``970`` MARC field through side effects.
    """
    def _is_schema_inspire_bai(id_, schema):
        return schema == 'INSPIRE BAI'

    def _is_schema_inspire_id(id_, schema):
        return schema == 'INSPIRE ID'

    def _is_schema_spires(id_, schema):
        return schema == 'SPIRES'

    def _is_schema_linkedin(id, schema):
        return schema == 'LINKEDIN'

    def _is_schema_twitter(id, schema):
        return schema == 'TWITTER'

    id_ = value.get('value')
    schema = value.get('schema')

    if _is_schema_spires(id_, schema):
        self.setdefault('970', []).append({'a': id_})
    elif _is_schema_linkedin(id_, schema):
        self.setdefault('8564', []).append(
            {
                'u': u'https://www.linkedin.com/in/{id}'.format(id=quote_url(id_)),
                'y': 'LINKEDIN',
            }
        )
    elif _is_schema_twitter(id_, schema):
        self.setdefault('8564', []).append(
            {
                'u': u'https://twitter.com/{id}'.format(id=id_),
                'y': 'TWITTER',
            }
        )
    elif _is_schema_inspire_id(id_, schema):
        return {
            'a': id_,
            '9': 'INSPIRE',
        }
    elif _is_schema_inspire_bai(id_, schema):
        return {
            'a': id_,
            '9': 'BAI',
        }
    else:
        return {
            'a': id_,
            '9': schema,
        }


@hepnames.over('name', '^100..')
def name(self, key, value):
    """Populate the ``name`` key.

    Also populates the ``status``, ``birth_date`` and ``death_date`` keys through side effects.
    """
    def _get_title(value):
        c_value = force_single_element(value.get('c', ''))
        if c_value != 'title (e.g. Sir)':
            return c_value

    def _get_value(value):
        a_value = force_single_element(value.get('a', ''))
        q_value = force_single_element(value.get('q', ''))
        return a_value or normalize_name(q_value)

    if value.get('d'):
        dates = value['d']
        try:
            self['death_date'] = normalize_date(dates)
        except ValueError:
            dates = dates.split(' - ')
            if len(dates) == 1:
                dates = dates[0].split('-')
            self['birth_date'] = normalize_date(dates[0])
            self['death_date'] = normalize_date(dates[1])

    self['status'] = force_single_element(value.get('g', '')).lower()

    return {
        'numeration': force_single_element(value.get('b', '')),
        'preferred_name': force_single_element(value.get('q', '')),
        'title': _get_title(value),
        'value': _get_value(value),
    }


@hepnames2marc.over('100', '^name$')
def name2marc(self, key, value):
    """Populates the ``100`` field.

    Also populates the ``400``, ``880``, and ``667`` fields through side
    effects.
    """
    result = self.get('100', {})

    result['a'] = value.get('value')
    result['b'] = value.get('numeration')
    result['c'] = value.get('title')
    result['q'] = value.get('preferred_name')

    if 'name_variants' in value:
        self['400'] = [{'a': el} for el in value['name_variants']]
    if 'native_names' in value:
        self['880'] = [{'a': el} for el in value['native_names']]
    if 'previous_names' in value:
        prev_names = [
            {'a': u'Formerly {}'.format(prev_name)}
            for prev_name in value['previous_names']
        ]
        self['667'] = prev_names

    return result


@hepnames2marc.over('100', '^status$')
def status2marc(self, key, value):
    result = self.get('100', {})

    result['g'] = value

    return result


@hepnames.over('positions', '^371..')
@utils.for_each_value
def positions(self, key, value):
    """Populate the positions field.

    Also populates the email_addresses field by side effect.
    """
    email_addresses = self.get("email_addresses", [])
    current = None
    record = None

    recid_or_status = force_list(value.get('z'))
    for el in recid_or_status:
        if el.lower() == 'current':
            current = True if value.get('a') else None
        else:
            record = get_record_ref(maybe_int(el), 'institutions')

    rank = normalize_rank(value.get('r'))

    current_email_addresses = force_list(value.get('m'))
    non_current_email_addresses = force_list(value.get('o'))

    email_addresses.extend({
        'value': address,
        'current': True,
    } for address in current_email_addresses)
    email_addresses.extend({
        'value': address,
        'current': False,
    } for address in non_current_email_addresses)

    self['email_addresses'] = email_addresses

    if 'a' not in value:
        return None

    return {
        'institution': value['a'],
        'record': record,
        'curated_relation': True if record is not None else None,
        'rank': rank,
        'start_date': normalize_date(value.get('s')),
        'end_date': normalize_date(value.get('t')),
        'current': current,
    }


@hepnames2marc.over('371', '^positions$')
@utils.for_each_value
def positions2marc(self, key, value):
    def _get_r_value(value):
        RANK_MAP = {
            'JUNIOR': 'JUNIOR',
            'MASTER': 'MAS',
            'PHD': 'PHD',
            'POSTDOC': 'PD',
            'SENIOR': 'SENIOR',
            'STAFF': 'STAFF',
            'UNDERGRADUATE': 'UG',
            'VISITOR': 'VISITOR',
        }

        rank = value.get('rank')
        if rank:
            return RANK_MAP.get(rank)

    return {
        'a': value.get('institution'),
        'r': _get_r_value(value),
        's': value.get('start_date'),
        't': value.get('end_date'),
        'z': 'Current' if value.get('current') else None,
    }


@hepnames2marc.over('595', '^email_addresses$')
@utils.for_each_value
def email_addresses2marc(self, key, value):
    """Populate the 595 MARCXML field.

    Also populates the 371 field as a side effect.
    """
    m_or_o = 'm' if value.get('current') else 'o'
    element = {
        m_or_o: value.get('value')
    }

    if value.get('hidden'):
        return element
    else:
        self.setdefault('371', []).append(element)
        return None


@hepnames.over('email_addresses', '^595..')
def email_addresses595(self, key, value):
    """Populates the ``email_addresses`` field using the 595 MARCXML field.

    Also populates ``_private_notes`` as a side effect.
    """
    emails = self.get('email_addresses', [])

    if value.get('o'):
        emails.append({
            'value': value.get('o'),
            'current': False,
            'hidden': True,
        })

    if value.get('m'):
        emails.append({
            'value': value.get('m'),
            'current': True,
            'hidden': True,
        })

    notes = self.get('_private_notes', [])
    new_note = (
        {
            'source': value.get('9'),
            'value': _private_note,
        } for _private_note in force_list(value.get('a'))
    )
    notes.extend(new_note)
    self['_private_notes'] = notes

    return emails


@hepnames.over('name', '^400..')
def name_variants(self, key, value):

    name_item = self.get('name', {})
    name_variants_list = name_item.get('name_variants', [])

    name_variants_list.extend(force_list(value.get('a')))
    name_item['name_variants'] = name_variants_list

    return name_item


@hepnames.over('arxiv_categories', '^65017')
def arxiv_categories(self, key, value):
    """Populate the ``arxiv_categories`` key.

    Also populates the ``inspire_categories`` key through side effects.
    """
    def _is_arxiv(category):
        return category in valid_arxiv_categories()

    def _is_inspire(category):
        schema = load_schema('elements/inspire_field')
        valid_inspire_categories = schema['properties']['term']['enum']

        return category in valid_inspire_categories

    def _normalize(a_value):
        for category in valid_arxiv_categories():
            if a_value.lower() == category.lower():
                return normalize_arxiv_category(category)

        schema = load_schema('elements/inspire_field')
        valid_inspire_categories = schema['properties']['term']['enum']

        for category in valid_inspire_categories:
            if a_value.lower() == category.lower():
                return category

        field_codes_to_inspire_categories = {
            'a': 'Astrophysics',
            'b': 'Accelerators',
            'c': 'Computing',
            'e': 'Experiment-HEP',
            'g': 'Gravitation and Cosmology',
            'i': 'Instrumentation',
            'l': 'Lattice',
            'm': 'Math and Math Physics',
            'n': 'Theory-Nucl',
            'o': 'Other',
            'p': 'Phenomenology-HEP',
            'q': 'General Physics',
            't': 'Theory-HEP',
            'x': 'Experiment-Nucl',
        }

        return field_codes_to_inspire_categories.get(a_value.lower())

    arxiv_categories = self.get('arxiv_categories', [])
    inspire_categories = self.get('inspire_categories', [])

    for value in force_list(value):
        for a_value in force_list(value.get('a')):
            normalized_a_value = _normalize(a_value)

            if _is_arxiv(normalized_a_value):
                arxiv_categories.append(normalized_a_value)
            elif _is_inspire(normalized_a_value):
                inspire_categories.append({'term': normalized_a_value})

    self['inspire_categories'] = inspire_categories
    return arxiv_categories


@hepnames2marc.over('65017', '^arxiv_categories$')
@utils.for_each_value
def arxiv_categories2marc(self, key, value):
    return {
        '2': 'arXiv',
        'a': value,
    }


@hepnames2marc.over('65017', '^inspire_categories$')
@utils.for_each_value
def inspire_categories2marc(self, key, value):
    return {
        '2': 'INSPIRE',
        'a': value.get('term'),
    }


@hepnames.over('public_notes', '^667..')
@utils.for_each_value
def _public_notes(self, key, value):
    if 'Formerly' in value.get('a'):
        name_item = self.get('name', {})
        previous_names_list = name_item.get('previous_names', [])
        previous_name = value.get('a').replace('Formerly ', '')
        previous_names_list.extend(force_list(previous_name))
        self.setdefault('name', {})['previous_names'] = previous_names_list
    else:
        return {
            'source': value.get('9'),
            'value': value.get('a'),
        }


@hepnames2marc.over('667', '^public_notes$')
@utils.for_each_value
def _public_notes2marc(self, key, value):
    return {
        'a': value.get('value'),
        '9': value.get('source'),
    }


@hepnames2marc.over('100', '^(birth_date|death_date)$')
def birth_and_death_date2marc(self, key, value):
    """Populate the ``100__d`` MARC field, which includes the birth and the death date.

    By not using the decorator ```for_each_value```, the values of the fields
    ```birth_date``` and ```death_date``` are both added to ```values``` as a list.
    """
    name_field = self.get('100', {})

    if 'd' in name_field:
        if int(name_field['d'].split('-')[0]) > int(value.split('-')[0]):
            dates_field = ' - '.join([value, name_field['d']])
        else:
            dates_field = ' - '.join([name_field['d'], value])
    else:
        dates_field = value

    name_field['d'] = dates_field

    return name_field


@hepnames.over('awards', '^678..')
@utils.for_each_value
def awards(self, key, value):
    award = AWARD_YEAR.sub('', value.get('a')).strip()
    year_match = AWARD_YEAR.search(value.get('a'))
    if year_match:
        year = int(year_match.group('year'))
    else:
        year = None

    return {
        'name': award,
        'url': value.get('u'),
        'year': year,
    }


@hepnames2marc.over('678', '^awards$')
@utils.for_each_value
def awards2marc(self, key, value):
    return {
        'a': ' '.join([value.get('name', ''), str(value.get('year', ''))]).strip(),
        'u': value.get('url')
    }


@hepnames.over('project_membership', '^693..')
def project_membership(self, key, values):
    def _get_json_experiments(marc_dict):
        start_year = marc_dict.get('s')
        end_year = marc_dict.get('d')

        names = force_list(marc_dict.get('e'))
        recids = force_list(marc_dict.get('0'))
        name_recs = zip(names, recids or [None] * len(names))

        for name, recid in name_recs:
            record = get_record_ref(recid, 'experiments')
            yield {
                'curated_relation': record is not None,
                'current': (
                    True if marc_dict.get('z', '').lower() == 'current'
                    else False
                ),
                'end_date': end_year,
                'name': name,
                'record': record,
                'start_date': start_year,
            }

    values = force_list(values)
    json_experiments = self.get('project_membership', [])
    for experiment in values:
        if experiment:
            json_experiments.extend(_get_json_experiments(experiment))

    return json_experiments


@hepnames2marc.over('693', '^project_membership$')
def project_membership2marc(self, key, values):
    def _get_marc_experiment(json_dict):
        marc = {
            'e': json_dict.get('name'),
            's': json_dict.get('start_date'),
            'd': json_dict.get('end_date'),
        }
        status = 'current' if json_dict.get('current') else None
        if status:
            marc['z'] = status
        recid = get_recid_from_ref(json_dict.get('record', None))
        if recid:
            marc['0'] = recid
        return marc

    marc_experiments = self.get('693', [])
    values = force_list(values)
    for experiment in values:
        if experiment:
            marc_experiments.append(_get_marc_experiment(experiment))

    return marc_experiments


@hepnames.over('advisors', '^701..')
@utils.for_each_value
def advisors(self, key, value):
    DEGREE_TYPES_MAP = {
        'Bachelor': 'bachelor',
        'UG': 'bachelor',
        'MAS': 'master',
        'master': 'master',
        'Master': 'master',
        'PhD': 'phd',
        'PHD': 'phd',
    }

    def _get_id_schema(id_):
        if id_.lower().startswith('inspire-'):
            return 'INSPIRE ID'
        else:  # assuming ORCID
            return 'ORCID'

    _degree_type = force_single_element(value.get('g'))
    degree_type = DEGREE_TYPES_MAP.get(_degree_type, 'other')

    recid = force_single_element(value.get('x'))
    record = get_record_ref(recid, 'authors')

    ids = [{
        'schema': _get_id_schema(id_),
        'value': id_,
    } for id_ in force_list(value.get('i'))]

    return {
        'name': value.get('a'),
        'degree_type': degree_type,
        'ids': ids,
        'record': record,
        'curated_relation': value.get('y') == '1' if record else None
    }


@hepnames2marc.over('701', '^advisors$')
@utils.for_each_value
def advisors2marc(self, key, value):
    allowed_ids = ('INSPIRE ID', 'ORCID')
    ids = [id_['value'] for id_ in value.get('ids', []) if id_['schema'] in allowed_ids]

    return {
        'a': value.get('name'),
        'g': value.get('degree_type'),
        'i': ids,
    }


@hepnames.over('urls', '^8564.')
@utils.for_each_value
def urls(self, key, value):
    """Populate the ``url`` key.

    Also populates the ``ids`` key through side effects.
    """
    description = force_single_element(value.get('y'))
    url = value.get('u')

    linkedin_match = LINKEDIN_URL.match(url)
    twitter_match = TWITTER_URL.match(url)
    wikipedia_match = WIKIPEDIA_URL.match(url)
    if linkedin_match:
        self.setdefault('ids', []).append(
            {
                'schema': 'LINKEDIN',
                'value': unquote_url(linkedin_match.group('page')),
            }
        )
    elif twitter_match:
        self.setdefault('ids', []).append(
            {
                'schema': 'TWITTER',
                'value': twitter_match.group('handle'),
            }
        )
    elif wikipedia_match:
        lang = wikipedia_match.group('lang')
        page = unquote_url(wikipedia_match.group('page'))
        if lang != 'en':
            page = ':'.join([lang, page])
        self.setdefault('ids', []).append(
            {
                'schema': 'WIKIPEDIA',
                'value': page,
            }
        )
    else:
        return {
            'description': description,
            'value': url,
        }


@hepnames.over('name', '^880..')
def native_name(self, key, value):
    name = self.get('name', {})
    name.setdefault('native_names', []).append(value.get('a'))
    return name


@hepnames.over('new_record', '^970..')
def new_record(self, key, value):
    """Populate the ``new_record`` key.

    Also populates the ``ids`` key through side effects.
    """
    new_record = self.get('new_record', {})
    ids = self.get('ids', [])

    for value in force_list(value):
        for id_ in force_list(value.get('a')):
            ids.append({
                'schema': 'SPIRES',
                'value': id_,
            })

        new_recid = force_single_element(value.get('d', ''))
        if new_recid:
            new_record = get_record_ref(new_recid, 'authors')

    self['ids'] = ids
    return new_record


@hepnames.over('deleted', '^980..')
def deleted(self, key, value):
    """Populate the ``deleted`` key.

    Also populates the ``stub`` key through side effects.
    """
    def _is_deleted(value):
        return force_single_element(value.get('c', '')).upper() == 'DELETED'

    def _is_stub(value):
        return not (force_single_element(value.get('a', '')).upper() == 'USEFUL')

    deleted = self.get('deleted')
    stub = self.get('stub')

    for value in force_list(value):
        deleted = not deleted and _is_deleted(value)
        stub = not stub and _is_stub(value)

    self['stub'] = stub
    return deleted


@hepnames2marc.over('980', '^_collections$')
@utils.for_each_value
def _collections2marc(self, key, value):
    if value == 'Authors':
        return {'a': 'HEPNAMES'}


@hepnames2marc.over('980', '^deleted$')
@utils.for_each_value
def deleted2marc(self, key, value):
    if value:
        return {'c': 'DELETED'}


@hepnames2marc.over('980', '^stub$')
@utils.for_each_value
def stub2marc(self, key, value):
    if not value:
        return {'a': 'USEFUL'}



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""DoJSON model and rules for experiments."""

from __future__ import absolute_import, division, print_function

from . import rules  # noqa: F401
from .model import experiments  # noqa: F401



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""DoJSON rules for jobs."""

from __future__ import absolute_import, division, print_function

import re

from dojson import utils
from six.moves import zip_longest

from inspire_utils.helpers import force_list
from inspire_utils.name import normalize_name
from inspire_schemas.utils import (
    normalize_arxiv_category,
    sanitize_html,
)

from .model import jobs
from ..utils import (
    force_single_element,
    get_record_ref,
    normalize_rank,
    normalize_date_aggressively,
)


COMMA_OR_SLASH = re.compile(r'\s*[/,]\s*')
EXTERNAL_ID = re.compile(r'\s*\((.*)\)\s*$')


@jobs.over('deadline_date', '^046..')
def deadline_date(self, key, value):
    el = force_single_element(value)

    deadline_date = force_single_element(el.get('i'))
    if not deadline_date or deadline_date == '0000':
        return '3000'
    return normalize_date_aggressively(deadline_date)


@jobs.over('contact_details', '^270..')
def contact_details(self, key, value):
    """Populate the ``contact_details`` key.

    Also populates the ``reference_letters`` key through side effects.
    """
    contact_details = self.get('contact_details', [])
    reference_letters = self.get('reference_letters', {})

    emails = force_list(value.get('m'))
    names = force_list(value.get('p'))
    if len(names) == 1 and len(emails) > 1:
        names = [names[0] for _ in emails]
    values_o = force_list(value.get('o'))

    contact_details.extend({
        'name': normalize_name(name),
        'email': email,
    } for (name, email) in zip_longest(names, emails))

    for value_o in values_o:
        if '@' in value_o:
            reference_letters.setdefault('emails', []).append(value_o)
        else:
            reference_letters.setdefault('urls', []).append({
                'value': value_o,
            })

    self['reference_letters'] = reference_letters
    return contact_details


@jobs.over('regions', '^043..')
def regions(self, key, value):
    REGIONS_MAP = {
        'AF': 'Africa',
        'Africa': 'Africa',
        'Asia': 'Asia',
        'Australia': 'Australasia',
        'Australasia': 'Australasia',
        'eu': 'Europe',
        'Europe': 'Europe',
        'Middle East': 'Middle East',
        'na': 'North America',
        'United States': 'North America',
        'Noth America': 'North America',
        'North America': 'North America',
        'North Americsa': 'North America',
        'South America': 'South America',
    }

    result = []

    for el in force_list(value.get('a')):
        for region in COMMA_OR_SLASH.split(el):
            result.append(REGIONS_MAP.get(region))

    return result


@jobs.over('accelerator_experiments', '^693..')
@utils.for_each_value
def accelerator_experiments(self, key, value):
    legacy_name = value.get('e')
    recid = value.get('0')
    record = get_record_ref(recid, 'experiments')

    return {
        'curated_relation': record is not None,
        'legacy_name': legacy_name,
        'record': record
    }


@jobs.over('institutions', '^110..')
def institutions(self, key, value):
    institutions = self.get('institutions', [])

    a_values = force_list(value.get('a'))
    z_values = force_list(value.get('z'))

    # XXX: we zip only when they have the same length, otherwise
    #      we might match a value with the wrong recid.
    if len(a_values) == len(z_values):
        for a_value, z_value in zip(a_values, z_values):
            record = get_record_ref(z_value, 'institutions')
            institutions.append({
                'curated_relation': record is not None,
                'value': a_value,
                'record': record,
            })
    else:
        for a_value in a_values:
            institutions.append({
                'curated_relation': False,
                'value': a_value,
            })

    return institutions


@jobs.over('description', '^520..')
def description(self, key, value):
    return sanitize_html(value.get('a', '')).strip()


@jobs.over('position', '^245..')
def position(self, key, value):
    """Populate the ``position`` key.

    Also populates the ``external_job_identifier`` key through side-effects.
    """
    try:
        external_id = EXTERNAL_ID.search(value.get('a', '')).group(1)
    except AttributeError:
        external_id = None

    position = EXTERNAL_ID.sub('', value.get('a', ''))

    self['external_job_identifier'] = external_id
    return position


@jobs.over('ranks', '^656..')
@utils.flatten
@utils.for_each_value
def ranks(self, key, value):
    """Populate the ``ranks`` key."""
    return [normalize_rank(el) for el in force_list(value.get('a'))]


@jobs.over('arxiv_categories', '^65017')
@utils.for_each_value
def arxiv_categories(self, key, value):
    category = value.get('a', '')

    if category.lower() == 'physics-other':
        return None
    elif category.lower() == 'physics.acc-phys':
        return 'physics.acc-ph'

    return normalize_arxiv_category(value.get('a'))


@jobs.over('status', '^980..')
def status(self, key, value):
    """Populate the ``status`` key. Also populates the ``deleted`` key through
    side-effects.
    """
    collection_to_status = {
        'JOBHIDDEN': 'closed',
        'JOB': 'open',
    }

    status = self.get('status')
    if value.get('c', '').upper() == 'DELETED':
        self['deleted'] = True

    status = collection_to_status.get(value.get('a', '').upper(), status)
    return status
